                                                        /*      CSSTRUCT.H  */
#ifndef CSOUND_CSSTRUCT_H
#define CSOUND_CSSTRUCT_H

/* csstruct.h -- written by Istvan Varga, Feb 2003 */

#ifdef __cplusplus
extern "C" {
#endif

/* for MYFLT */
#include "Csound/sysdep.h"

/* --------------------------- public data types --------------------------- */

/* opcode function: void name(void *p) */

typedef void    (*SUBR)(void*);

/* Structures and macros for function tables. */

/* maximum string length (do not modify this) */
#define SSTRSIZ 200

#define MAXLEN      0x1000000L
#define FMAXLEN     ((MYFLT) (MAXLEN))
#define PHMASK      0x0FFFFFFL
#define PFRAC(x)    ((MYFLT) ((x) & ftp->lomask) * ftp->lodiv)
#define MAXPOS      0x7FFFFFFFL

#define LOBITS      10
#define LOFACT      1024
/* LOSCAL is 1 / LOFACT as MYFLT */
#define LOSCAL      FL(0.0009765625)
#define LOMASK      1023

/* parameters of GEN01 sound file */

typedef struct {
    MYFLT   gen01;
    MYFLT   ifilno;
    MYFLT   iskptim;
    MYFLT   iformat;
    MYFLT   channel;
    MYFLT   sample_rate;
    char    strarg[SSTRSIZ];
} GEN01ARGS;

/* function table */

typedef struct {
    long    flen;               /* table length (not incl. guard point)   */
    long    lenmask;            /* flen - 1 (for wrapping table index)    */
    long    lobits;             /* log2(MAXLEN / flen)                    */
    long    lomask;             /* (MAXLEN / flen) - 1                    */
    MYFLT   lodiv;              /* flen / MAXLEN (see also PFRAC macro)   */

    MYFLT   cvtbas, cpscvt;     /* misc. data for GEN01 sound files       */
    short   loopmode1;
    short   loopmode2;
    long    begin1, end1;       /* all these in ..  */
    long    begin2, end2;
    long    soundend, flenfrms; /* .. sample frames */
    long    nchanls;
    long    fno;
    GEN01ARGS   gen01args;

    MYFLT   ftable[1];          /* table data (flen + 1 floats)           */
} FUNC;

/* some commonly used structures */

/* files opened by opcodes (CSOUNDAPI.OpenFile()) */

typedef struct fdch {
    struct fdch *nxtchp;        /* pointer to next structure in chain     */
    int     fd;                 /* file descriptor as returned by open()  */
} FDCH;

/* for auxiliary memory allocation by opcodes (CSOUNDAPI.AuxAlloc()) */

typedef struct auxch {
    char    *auxp;              /* pointer to allocated memory area       */
    char    *endp;              /* pointer to last byte + 1               */
    long    size;               /* number of bytes allocated              */
} AUXCH;

/* file loaded into memory (CSOUNDAPI.LoadFile()) */

typedef struct MEMFIL {
    char    filename[256];      /* file name (made larger RWD)            */
    char    *beginp;            /* pointer to memory area storing file    */
    char    *endp;              /* pointer to last byte + 1               */
    long    length;             /* file length in bytes                   */
    struct MEMFIL *next;        /* pointer to next structure in chain     */
} MEMFIL;

/* the following four structures are for MIDI (FIXME: find out */
/* what these are exactly ...)                                 */

typedef struct monblk {
        short  pch;
        struct monblk *prv;
} MONPCH;

typedef struct {
        int     notnum[4];
} DPEXCL;

typedef struct {
        DPEXCL dpexcl[8];
        int    exclset[75];     /* for keys 25-99 */
} DPARM;

typedef struct dklst {
        struct dklst *nxtlst;
        long   pgmno;
        MYFLT  keylst[1];       /* cnt + keynos */
} DKLST;

/* MIDI channel block */

typedef struct mchnblk {
    short   pgmno;              /* program number (0 to 127)              */
    short   insno;              /* instrument number (1 to maxinsno)      */
    short   RegParNo;           /* ??? (unknown)                          */
    short   mono;               /* ??? (unknown)                          */
    MONPCH  *monobas;           /* ??? (unknown)                          */
    MONPCH  *monocur;           /* ??? (unknown)                          */
#ifdef CSOUND_API_HEADERS
    void    *kinsptr[128];      /* unused (internal Csound variable)      */
    void    *ksusptr[128];      /* unused (internal Csound variable)      */
#else
    struct insds *kinsptr[128]; /* array of active keys, indexed by pitch */
    struct insds *ksusptr[128]; /* array of sustaining keys by pitch      */
#endif
    MYFLT   polyaft[128];       /* polyphonic pressure, indexed by pitch  */
    MYFLT   ctl_val[128];       /* controller values for this channel,    */
                                /*  with GS vib_rate, stored in c102-c109 */
                                /* N.B. controller and aftertouch values  */
                                /*   are in the range 0.0 to 127.0        */
    short   ksuscnt;            /* number of sustaining keys              */
    short   sustaining;         /* sustain pedal on ? (non-zero: yes)     */
    MYFLT   aftouch;            /* channel pressure (range: 0.0 to 127.0) */
    MYFLT   pchbend;            /* pitch bend (range: -1.0 to 1.0)        */
    DKLST   *klists;            /* chain of dpgm keylists (???)           */
    DPARM   *dparms;            /* drumset params                         */
    int     dpmsb;              /* ??? (unknown)                          */
    int     dplsb;              /* ??? (unknown)                          */
} MCHNBLK;

/* Magic value for identifying string arguments. Do not change this. */
#define SSTRCOD ((MYFLT) 1045763)
/* maximum number of opcode arguments */
#define VARGMAX  1001

#define  ONEPT          1.021975               /* A440 tuning factor */
#define  LOG10D20       0.11512925             /* for db to ampfac   */
#define  DV32768        FL(0.000030517578125)

/* Standard audio encoding types */
#define AE_CHAR   0x101         /* Signed 8 bit             */
#define AE_ALAW   0x102         /* a-law (not supported)    */
#define AE_ULAW   0x103         /* u-law                    */
#define AE_SHORT  0x104         /* 16 bit signed integer    */
#define AE_LONG   0x105         /* 32 bit signed integer    */
#define AE_FLOAT  0x106         /* 32 bit float             */
#define AE_UNCH   0x107         /* Unsigned 8 bit           */
#define AE_24INT  0x108         /* 24 bit signed integer    */
#define AE_DOUBLE 0x109         /* 64 bit float             */
#define AE_LAST   AE_DOUBLE     /* current last audio encoding value */

/* Audio file types */
#define TYP_IRCAM 0
#define TYP_AIFF  1
#define TYP_WAV   2
#define TYP_AIFC  3

typedef struct AUDIOPARMS_S {
    int     deviceNumber;       /* device number (default: 1024)            */
    int     nChannels;          /* number of channels (1: mono, etc.)       */
    float   sampleRate;         /* sample rate in Hz                        */
    float   __dummy_01;         /* unused                                   */
    int     sampleFormat;       /* sample format (AE_CHAR etc., see above)  */
    int     sampleSize;         /* sample size (e.g. 4 for AE_LONG)         */
    int     swBufSize;          /* software buffer size (-b) in samples     */
    int     hwBufSize;          /* hardware buffer size (-B) in samples     */
    char    *deviceName;        /* device name (default: NULL)              */
    void    *bufPtr;            /* pointer to audio I/O buffer              */
    int     schedPriority;      /* scheduling priority (system dependent)   */
    int     sleepTime;          /* sleep time (system dependent)            */
    int     userParam1;         /* special parameter 1 (system dependent)   */
    int     userParam2;         /* special parameter 2 (system dependent)   */
} AUDIOPARMS;


#ifdef CSOUND_API_HEADERS

/* opcode entry structure for use by plugin opcodes */

typedef struct oentry {
    char    *opname;            /* opcode name                            */
    unsigned short  dsblksiz;   /* opcode structure size                  */
    unsigned short  thread;     /* thread types (1: i-, 2: k-, 4: a-rate) */
    char    *outypes;           /* output arg types                       */
    char    *intypes;           /* input arg types                        */
    SUBR    iopadr;             /* initialisation function pointer        */
    SUBR    kopadr;             /* k-rate function pointer                */
    SUBR    aopadr;             /* a-rate function pointer                */
/*
 * Deinitialization function pointer;
 * if not null, called during cleanup on each opcode instance;
 * useful for deallocating memory or other resources managed by the opcode.
 */
    SUBR    dopadr;

    void    *dummy1;    /* unused (internal Csound variable) */
    int     dummy2;     /* unused (internal Csound variable) */
} OENTRY;

typedef void    INSDS;          /* stubs */
typedef void    opcodelist;

typedef struct opds {
        struct opds     *nxti;  /* Next opcode in init-time chain            */
        struct opds     *nxtp;  /* Next opcode in perf-time chain            */
        SUBR    iopadr;         /* Initialization (i-time) function pointer  */
        SUBR    opadr;          /* Perf-time (k- or a-rate) function pointer */
        void    *dummy1;        /* unused (internal Csound variable)         */
        void    *dummy2;        /* unused (internal Csound variable)         */
        void    *dummy3;        /* unused (internal Csound variable)         */
        struct CSGLOBALS_S      /* Pointer to globals structure of           */
                *globals;       /*   owner Csound instance                   */
} OPDS;

/* for timers */

typedef struct RTCLOCK_S {
    unsigned long   starttime_real_high;
    unsigned long   starttime_real_low;
    double          real_time_to_seconds_scale;
    unsigned long   starttime_CPU_high;
    unsigned long   starttime_CPU_low;
    double          CPU_time_to_seconds_scale;
} RTCLOCK;

#else           /* CSOUND_API_HEADERS */

/* -- the following structures are only for internal use by Csound sources -- */

#include "timers.h"

typedef struct arglst {
        int     count;
        char    *arg[1];
} ARGLST;

typedef struct argoffs {
        int     count;
        int     indx[1];
} ARGOFFS;

/* Storage for parsed orchestra code, for each opcode in an INSTRTXT. */
typedef struct text {
        short   linenum;    /* Line num in orch file (currently buggy!)      */
        short   opnum;      /* Opcode index in opcodlst[]                    */
        char    *opcod;     /* Pointer to opcode name in global pool         */
        char    *strargs[4];/* (Unquoted) array of file names if opcode uses */
        ARGLST  *inlist;    /* Input args (pointer to item in name list)     */
        ARGLST  *outlist;
        ARGOFFS *inoffs;    /* Input args (index into list of values)        */
        ARGOFFS *outoffs;
        short   xincod;     /* Rate switch for multi-rate opcode functions   */
        short   xoutcod;    /* output rate switch (IV - Sep 1 2002)          */
        char    intype;     /* Type of first input argument (g,k,a,w etc)    */
        char    pftype;     /* Type of output argument (k, a etc)            */
} TEXT;

/* This struct is filled out by otran() at orch parse time.
   It is used as a template for instrument events. */
typedef struct instr {
        struct op * nxtop;          /* Linked list of instr opcodes          */
        TEXT    t;                  /* Text of instrument (same in nxtop)    */
        short   pmax, pextrab;      /* Arg count, size of data for all
                                       opcodes in instr                      */
        short   threads;            /* Opcode type (i/k/a). This value is a  */
                                    /* bitwise OR of OENTRY.thread of all    */
                                    /* opcodes in this instrument.           */
        short   muted;              /* 0: instrument is muted                */
        int     lclkcnt, lcldcnt;   /* Storage reqs for this instr           */
        int     lclwcnt, lclacnt;
        int     lclpcnt;
        int     lclfixed, optxtcount;
        long    localen;
        long    opdstot;            /* Total size of opds structs in instr   */
        long    *inslist;           /* Only used in parsing (?)              */
        MYFLT   *psetdata;          /* Used for pset opcode                  */
        struct insds * instance;    /* Chain of allocated instances of
                                       this instrument                       */
        struct insds * lst_instance, *act_instance;     /* IV - Oct 26 2002  */
        struct instr * nxtinstxt;   /* Next instrument in orch (num order)   */
        int     active;             /* To count activations for control      */
        int     maxalloc;
        MYFLT   cpuload;            /* % load this instrumemnt makes         */
        struct opcodinfo *opcode_info;  /* IV - Nov 10 2002 */
        char    *insname;               /* instrument name  */
} INSTRTXT;

/* A chain of TEXT structs. Note that this is identical with the first two
   members of struct INSTRTEXT, and is so typecast at various points in code. */
typedef struct op {
        struct op * nxtop;
        TEXT    t;
} OPTXT;

/* This struct holds the info for a concrete instrument event
   instance in performance. */
typedef struct insds {
        struct opds   *nxti;    /* Chain of init-time opcodes */
        struct opds   *nxtp;    /* Chain of performance-time opcodes */
        void    *opcod_iobufs;  /* IV - Jan 08 2003: user opcode I/O buffers */
        int     relesing;       /* Flag to indicate we're releasing, test */
                                /*   with release opcode (IV: changed to int) */
        struct insds
                *nxtinstance;   /* Next allocated instance */
        struct insds
                *prvinstance;   /* Previous allocated instance            */
        struct insds  *nxtact;  /* Next in list of active instruments     */
        struct insds  *prvact;  /* Previous in list of active instruments */
        struct insds  *nxtoff;  /* Next instrument to terminate           */
        FDCH    fdch;       /* Chain of files used by opcodes in this instr */
        void    *aux_heap;  /* Extra memory used by opcodes in this instr   */
        MCHNBLK *m_chnbp;       /* MIDI note info block if event started from */
                                /*   MIDI (otherwise NULL)                */
        short   m_pitch;        /* MIDI pitch, for simple access          */
        short   m_veloc;        /* ...ditto velocity                      */
        int     xtratim;        /* Extra release time requested with xtratim  */
                                /*   opcode (IV - Jan 08 2003: use int type)  */
                                /* (N.B. it is always in global_onedkr units) */
        short   insno;          /* Instrument number                      */
        short   actflg;         /* Set if instr instance is active        */
                                /*   (performing)                         */
        MYFLT   offbet;         /* Time to turn off event, in score beats */
        MYFLT   offtim;         /* Time to turn off event, in seconds     */
                                /*   (negative on indef/tie)              */
        void    *opcod_deact;   /* IV - Oct 24 2002 */
        void    *subins_deact;
        /* IV - Feb 22 2003 */
        void    *csound;        /* pointer to Csound engine and API       */
        void    *csglobals;     /* pointer to standard global variables   */
        struct opds     *nxtd;  /* chain of deactivation time opcodes     */
        struct opds     *nxts;  /* chain of section end time opcodes      */
        MYFLT   p0;             /* Copy of required p-field values for    */
        MYFLT   p1;             /*   quick access                         */
        MYFLT   p2;
        MYFLT   p3;
} INSDS;

/* Extensions to the OPDS structure. */

typedef struct opdsx {
        struct opds * nxtd;     /* Next opcode in deactivation time chain   */
        struct opds * nxts;     /* Next opcode in section end time chain    */
/*
 * Deinitialization function pointer;
 * if not null, called during cleanup on each opcode instance;
 * useful for deallocating memory or other resources managed by the opcode.
 */
        SUBR    dopadr;         /* Deinitialization function pointer        */
} OPDSX;

/* This struct holds the info for one opcode in a concrete
   instrument instance in performance. */
typedef struct opds {
        struct opds * nxti;     /* Next opcode in init-time chain */
        struct opds * nxtp;     /* Next opcode in perf-time chain */
        SUBR    iopadr;         /* Initialization (i-time) function pointer */
        SUBR    opadr;          /* Perf-time (k- or a-rate) function pointer */
/* Pointer to OPDS extensions (can be NULL if the opcode does not depend on */
/* any of the special features defined in OPDSX). OPDSX allows extending    */
/* opcode functionality without breaking plugins that depend on the layout  */
/* of OPDS.                                                                 */
        OPDSX   *opds_ex;       /* Extensions to the OPDS structure. */
        OPTXT   *optext;        /* Orch file template part for this opcode */
        INSDS   *insdshead;     /* Owner instrument instance data structure */
        struct CSGLOBALS_S      /* Pointer to globals structure of */
                *globals;       /*   owner Csound instance         */
} OPDS;

typedef struct oentry {
    /* ---- public data ---- */
        char    *opname;
        unsigned short  dsblksiz;
        unsigned short  thread;
        char    *outypes;
        char    *intypes;
        SUBR    iopadr;
        SUBR    kopadr;
        SUBR    aopadr;
/*
 * Deinitialization function pointer;
 * if not null, called during cleanup on each opcode instance;
 * useful for deallocating memory or other resources managed by the opcode.
 */
        SUBR    dopadr;
    /* ---- private data ---- */
        void    *useropinfo;    /* IV - Oct 12 2002: user opcode parameters */
        int     prvnum;         /* IV - Oct 31 2002 */
} OENTRY;

typedef struct opcodinfo {              /* IV - Oct 24 2002 */
        long    instno;
        char    *name, *intypes, *outtypes;
        short   inchns, outchns, perf_incnt, perf_outcnt;
        short   *in_ndx_list, *out_ndx_list;
        INSTRTXT  *ip;
        struct  opcodinfo *prv;
} OPCODINFO;

#endif          /* !CSOUND_API_HEADERS */

typedef struct lblblk {
        OPDS    h;
        OPDS    *prvi;
        OPDS    *prvp;
} LBLBLK;

#ifdef __cplusplus
}
#endif

#endif          /* CSOUND_CSSTRUCT_H */

